import ckan.plugins as p
import paste.script
import db
import logging

from ckan.lib.cli import CkanCommand

from ckan import model

log = logging.getLogger(__name__)

class TagmanagerCommands(CkanCommand):
	"""
	ckanext-tagmanager commands:
	Usage::
		paster tagmanager migrate
	"""
	summary = __doc__.split('\n')[0]
	usage = __doc__

	parser = paste.script.command.Command.standard_parser(verbose=True)
	parser.add_option('-c', '--config', dest='config',
		default='development.ini', help='Config file to use.')

	def command(self):
		if not len(self.args):
			print self.__doc__
			return

		cmd = self.args[0]
		self._load_config()

		if cmd == 'migrate':
			self._migrate()
		else:
			print self.__doc__

	def _migrate(self):
		if not db.tag_merge_suggestions_table.exists():
			db.tag_merge_suggestions_table.create()
			log.info('Tag Merge Suggestions table created')
			print 'Tag Merge Suggestions table created'
		else:
			log.warning('Tag Merge Suggestions table already exists')
			print 'Tag Merge Suggestions table already exists'




import vdm.sqlalchemy
from sqlalchemy.orm import relation, relationship
from sqlalchemy import types, Column, Table, ForeignKey, and_, UniqueConstraint, or_

import ckan.model.tag as _tag
import ckan.model.extension as _extension
import ckan.model.core as core
import ckan.model.meta as meta
import ckan.model.types as _types
import ckan.model.domain_object as domain_object
import ckan.model.activity as activity
import ckan  # this import is needed
import ckan.lib.dictization

__all__ = ['tag_merge_suggestions_table', 'TagMergeSuggestion']

tag_id_1 = Column('tag_id_1', types.UnicodeText, ForeignKey('tag.id',ondelete="CASCADE"))
tag_id_2 = Column('tag_id_2', types.UnicodeText, ForeignKey('tag.id',ondelete="CASCADE"))

tag_merge_suggestions_table = Table('tag_merge_suggestions', meta.metadata,
		Column('id', types.UnicodeText, primary_key=True, default=_types.make_uuid),
		tag_id_1,
		tag_id_2,
		Column('suggestion_type', types.Unicode)
)

#vdm.sqlalchemy.make_table_stateful(tag_semantictag_table)
# TODO: this has a composite primary key ...
#tag_semantictag_revision_table = core.make_revisioned_table(tag_semantictag_table)

class TagMergeSuggestion(domain_object.DomainObject):
	def __init__(self, tag_id_1, tag_id_2, suggestion_type):
		self.tag_id_1 = tag_id_1
		self.tag_id_2 = tag_id_2
		self.suggestion_type = suggestion_type

	def __repr__(self):
		return '<TagMergeSuggestion ' + self.tag_id_1 + ' ' + self.tag_id_2 + '>'   

	# not stateful so same as purge
	def delete(self):
		self.purge()
		return

	@classmethod
	def by_id(cls, id, autoflush=True):
		'''Return the suggestion with the given id, or None.

		:param id: the id of the tag merge suggestion
		:type id: string

		:returns: the suggestion with correspondent id or none
		:rtype: ckan.model.tag_merge_suggestion.TagMergeSuggestion

		'''
		query = meta.Session.query(TagMergeSuggestion).\
		filter(TagMergeSuggestion.id == id)

		return query.autoflush(autoflush).first()

	@classmethod
	def by_tag_id(cls, tag_id, autoflush=True):
		'''Return the suggestion with the given tag id, or None.

		:param tag_id: the id of the tag to look for suggestion
		:type tag_id: string

		:returns: the suggestions containing the specific tag id
		:rtype: ckan.model.tag.Tag # TODO check this

		'''
		query = meta.Session.query(TagMergeSuggestion).\
		filter(or_(TagMergeSuggestion.tag_id_1==tag_id,TagMergeSuggestion.tag_id_2==tag_id))

		query = query.autoflush(autoflush)
		tag_merge_suggestions = query.all()
		return tag_merge_suggestions

	@classmethod
	def by_type(cls, suggestion_type, limit=None, autoflush=True):
		'''Return the suggestion for a given type. Type options are: 0 (difference of capitals or special characters), 1 (levensthein distance smaller than 3) and 2 (synonyms).

		:param suggestion_type: suggestion_type (values 0 to 2)
		:type suggestion_type: integer (0-2)

		:returns: the suggestions objects
		:rtype: ckan.model.tag_merge_suggestion.TagMergeSuggestion

		'''

		if limit:
			query = meta.Session.query(TagMergeSuggestion).\
			filter(TagMergeSuggestion.suggestion_type == str(suggestion_type)).\
			limit(limit)
		else:
			query = meta.Session.query(TagMergeSuggestion).\
			filter(TagMergeSuggestion.suggestion_type == str(suggestion_type))

		query = query.autoflush(autoflush)
		tag_merge_suggestions = query.all()
		return tag_merge_suggestions

	@classmethod
	def all(cls,autoflush=True):
		'''Return the all suggestions in Database.

		:param suggestion_type: suggestion_type (values 0 to 2)
		:type suggestion_type: integer (0-2)

		:returns: the suggestions objects
		:rtype: ckan.model.tag_merge_suggestion.TagMergeSuggestion

		'''
		query = meta.Session.query(TagMergeSuggestion)
		return query.all()

#class TagSemanticTag(vdm.sqlalchemy.RevisionedObjectMixin,
#		vdm.sqlalchemy.StatefulObjectMixin,
#		domain_object.DomainObject):i


meta.mapper(TagMergeSuggestion, tag_merge_suggestions_table, properties={
	'tag_1': relation(_tag.Tag, foreign_keys=[tag_id_1]),
	'tag_2': relation(_tag.Tag, foreign_keys=[tag_id_2])
	}
	)


#meta.mapper(TagSemanticTag, tag_semantictag_table, properties={
#	'smtag':relation(_tag.Tag, backref='tag_semantictag_all',
#		cascade='none',
#		)
#	},
#	order_by=tag_semantictag_table.c.id,
##	extension=[vdm.sqlalchemy.Revisioner(tag_semantictag_revision_table),
##			   _extension.PluginMapperExtension(),
##			   ],
#	)




from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
	long_description = f.read()

setup(
	name='''ckanext-tagmanager''',

	# Versions should comply with PEP440.  For a discussion on single-sourcing
	# the version across setup.py and the project code, see
	# http://packaging.python.org/en/latest/tutorial.html#version
	version='0.0.2',

	description='''Tag management for CKAN portals''',
	long_description=long_description,

	# The project's main homepage.
	url='https://github.com/alantygel/ckanext-tagmanager',

	# Author details
	author='''Alan Tygel''',
	author_email='''alantygel@gmail.com''',

	# Choose your license
	license='AGPL',

	# See https://pypi.python.org/pypi?%3Aaction=list_classifiers
	classifiers=[
		# How mature is this project? Common values are
		# 3 - Alpha
		# 4 - Beta
		# 5 - Production/Stable
		'Development Status :: 4 - Beta',

		# Pick your license as you wish (should match "license" above)
		'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',

		# Specify the Python versions you support here. In particular, ensure
		# that you indicate whether you support Python 2, Python 3 or both.
		'Programming Language :: Python :: 2.6',
		'Programming Language :: Python :: 2.7',
	],


	# What does your project relate to?
	keywords='''CKAN tag , management, organization, tag merging, tag cleansing''',

	# You can just specify the packages manually here if your project is
	# simple. Or you can use find_packages().
	packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

	# List run-time dependencies here.  These will be installed by pip when your
	# project is installed. For an analysis of "install_requires" vs pip's
	# requirements files see:
	# https://packaging.python.org/en/latest/technical.html#install-requires-vs-requirements-files
	install_requires=[],

	# If there are data files included in your packages that need to be
	# installed, specify them here.  If using Python 2.6 or less, then these
	# have to be included in MANIFEST.in as well.
	include_package_data=True,
	package_data={
	},

	# Although 'package_data' is the preferred approach, in some case you may
	# need to place data files outside of your packages.
	# see http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
	# In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
	data_files=[],

	# To provide executable scripts, use entry points in preference to the
	# "scripts" keyword. Entry points provide cross-platform support and allow
	# pip to create the appropriate form of executable for the target platform.
	entry_points='''
		[ckan.plugins]
		tagmanager=ckanext.tagmanager.plugin:TagmanagerPlugin
 
	[paste.paster_command]
		tagmanager=ckanext.tagmanager.commands:TagmanagerCommands

   '''
)



import ckan.plugins as plugins
import ckan.plugins.toolkit as toolkit
import Levenshtein
from unidecode import unidecode
import db
import ckan.model as model
import ckan.model.meta as meta

def list_tags():
    return plugins.toolkit.get_action('tag_list')({},{'all_fields' : True})

def tag_show(id):
    return plugins.toolkit.get_action('tag_show')({},{'id' : id, 'include_datasets': True})

def get_name(id):
    return plugins.toolkit.get_action('tag_show')({},{'id' : id, 'include_datasets': False})['display_name']

def tags_stats():
	total_tags = len(model.Tag.all().all())

	total_taggings = len(meta.Session.query(model.PackageTag).all())

	return {'tags': total_tags, 'taggings': total_taggings}
    #tags_count = []
    #for t in range(0,len(tags)):
	#print tags[t]
	#tags_count[t] = plugins.toolkit.get_action('tag_show')({},{'id' : tags[t]['id']})

def tag_count(id):
    return plugins.toolkit.get_action('tag_show')({},{'id' : id, 'include_datasets': True})

def get_suggestions(suggestion_type=0, limit=None):
	return db.TagMergeSuggestion.by_type(suggestion_type, limit)

#def tags_merge_list_0(limit=None):
#	tags = list_tags()
#	T = len(tags)
#	merge_list = []
#	count = 0
#	if limit == None:
#		limit = 1000 #MAX_LIMIT
#	for t in range(0,T-1):
#		for s in range(t+1, T-1):
#			if unidecode(tags[s]['name'].lower()) == unidecode(tags[t]['name'].lower()):
#				merge_list.append([tags[s],tags[t]])
#				count += 1
#				if count > limit:
#					return merge_list

#	return merge_list

	

#def tags_merge_list_1(limit = None):
#	tags = list_tags()
#	T = len(tags)
#	merge_list = []
#	count = 0
#	if limit == None:
#		limit = 1000 #MAX_LIMIT
#	for t in range(0,T-1):
#		#check if there are numbers
#		stri = tags[t]['name'] 
#		if ([int(stri[i]) for i in range(0,len(stri)) if stri[i].isdigit()] == []) and (len(stri) > 3):
#			for s in range(t,T-1):
#				strj = tags[s]['name']
#				if (len(strj) > 3) and ([int(strj[i]) for i in range(0,len(strj)) if strj[i].isdigit()] == []):
#					if unidecode(tags[s]['name'].lower()) != unidecode(tags[t]['name'].lower()):
#						d = Levenshtein.distance(tags[t]['name'],tags[s]['name'])
#						if d < 3:
#							merge_list.append([tags[s],tags[t]])
#							count += 1
#							if count > limit:
#			   					return merge_list
#		
#	return merge_list

#def tags_merge_list_2(limit = None):
#	from nltk.corpus import wordnet as wn
#	tags = list_tags()
#	T = len(tags)
#	merge_list = []
#	count = 0
#	if limit == None:
#		limit = 1000 #MAX_LIMIT
#	for t in range(0,T-1):
#	#check if there are numbers
#		stri = tags[t]['name'] 
#		if ([int(stri[i]) for i in range(0,len(stri)) if stri[i].isdigit()] == []) and (len(stri) > 3):
#			syn1=wn.synsets(stri)
#			if syn1 != []:
#				for s in range(t,T-1):
#					strj = tags[s]['name']
#					if (len(strj) > 3) and ([int(strj[i]) for i in range(0,len(strj)) if strj[i].isdigit()] == []):
#						syn2=wn.synsets(strj)
#						if syn2 != []:
#							if unidecode(tags[s]['name'].lower()) != unidecode(tags[t]['name'].lower()):
#								if Levenshtein.distance(tags[t]['name'],tags[s]['name']) > 3:
#									b = max(syn2[i].wup_similarity(syn1[0]) for i in range(len(syn2)))
#									if b >= 1:
#										merge_list.append([tags[s],tags[t]])
#										count += 1
#										if count > limit:
#						   					return merge_list
#		
#	return merge_list

def has_suggestions(suggestion_type='all'):
	if suggestion_type == 'all':
		suggestions = db.TagMergeSuggestion.all();
	else: 
		suggestions = db.TagMergeSuggestion.by_type(suggestion_type);

	return len(suggestions)

class TagmanagerPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IConfigurer)

    # IConfigurer

    plugins.implements(plugins.IConfigurable)
    plugins.implements(plugins.IRoutes, inherit=True)
    #p.implements(p.IDomainObjectModification, inherit=True)
    #p.implements(p.IResourceUrlChange)
    plugins.implements(plugins.ITemplateHelpers)

    def configure(self, config):
        self.site_url = config.get('ckan.site_url')

    def before_map(self, map):
        tagmanager = 'ckanext.tagmanager.controller:TagmanagerController'

        map.connect('/tagmanager', 'tagmanager', controller=tagmanager, action='index')
        map.connect('/tagmanager/edit', controller=tagmanager, action='edit')
        map.connect('/tagmanager/index_process_suggestions', controller=tagmanager, action='index_process_suggestions')
        map.connect('/tagmanager/save_merge_suggestions', controller=tagmanager, action='save_merge_suggestions')
        map.connect('/tagmanager/merge_0', controller=tagmanager, action='merge_0')
        map.connect('/tagmanager/merge_1', controller=tagmanager, action='merge_1')
        map.connect('/tagmanager/merge_2', controller=tagmanager, action='merge_2')
        map.connect('/tagmanager/merge_form', controller=tagmanager, action='merge_form')
        map.connect('/tagmanager', controller=tagmanager, action='index')
        map.connect('/tagmanager/merge_confirm', controller=tagmanager, action='merge_confirm')
        map.connect('/tagmanager/merge', controller=tagmanager, action='merge')
        map.connect('/tagmanager/merge_do', controller=tagmanager, action='merge_do')
        map.connect('/tagmanager/delete_confirm', controller=tagmanager, action='delete_confirm')
        map.connect('/tagmanager/delete', controller=tagmanager, action='delete')
	return map

    def after_map(self, map):
        return map

    def update_config(self, config_):
        toolkit.add_template_directory(config_, 'templates')
        toolkit.add_public_directory(config_, 'public')
        toolkit.add_resource('fanstatic', 'tagmanager')

    def get_helpers(self):
	return {'tagmanager_list_tags':list_tags,'tagmanager_tag_show':tag_show, 'tagmanager_tags_stats': tags_stats, 'tagmanager_tag_count':tag_count, 'tagmanager_has_suggestions':has_suggestions, 'tagmanager_get_suggestions': get_suggestions, 'tagmanager_get_name':get_name}



import ckan.plugins.toolkit as toolkit
import ckan.lib.helpers as helpers

import ckan.plugins as p
from ckan.lib.base import BaseController, response, request
import json 
import db

import Levenshtein
from unidecode import unidecode


c = p.toolkit.c
render = p.toolkit.render

class TagmanagerController(BaseController):

	#index function for display form to load datasets for managing their relations
	def index(self):
		return render('tagmanager/index.html')

	def edit(self):
		return render('tagmanager/edit.html')

	def index_process_suggestions(self):
		return render('tagmanager/index_process_suggestions.html')	

	def merge_0(self):
		return render('tagmanager/index_merge_0.html')	

	def merge_1(self):
		return render('tagmanager/index_merge_1.html')	
   
	def merge_2(self):
		return render('tagmanager/index_merge_2.html')	

	def merge_form(self):
		return render('tagmanager/merge_form.html')	

	def merge_confirm(self):
		return render('tagmanager/merge_confirm.html')	

	def delete_confirm(self):
		return render('tagmanager/delete_confirm.html')	

	def delete(self):
		p.toolkit.get_action('tag_delete')({},{'id': request.POST['tag']})
		return render('tagmanager/index.html')	

	def merge(self):
		"assign all elements tagged with tag2 with tag1; delete tag2"

		tag2_datasets = p.toolkit.get_action('tag_show')({},{'id' : request.POST['tag2'], 'include_datasets': True})

		for ds in tag2_datasets['packages']:
			dataset = p.toolkit.get_action('package_show')({},{'id': ds['id'] })
			dataset['tags'].append(p.toolkit.get_action('tag_show')({},{'id':request.POST['tag1']}))
			p.toolkit.get_action('package_update')({},dataset)
		   
		p.toolkit.get_action('tag_delete')({},{'id': request.POST['tag2']})
	
		#p.toolkit.redirect_to(controller='tagmanager', action='index')

		return render('tagmanager/index.html')

	def merge_do(self):
		"assign all elements tagged with tag2 with tag1; delete tag2"

		merges = request.POST.getall('merge')
		for m in merges:
			self.merge_from_suggestion(m,request.POST["select_" + m])

		return render('tagmanager/index.html')

	def merge_from_suggestion(self, merge_id, tag_maintain):
		"assign all elements tagged with tag2 with tag1; delete tag2"
		
		merge_object = db.TagMergeSuggestion.by_id(merge_id)

		if merge_object.tag_id_1 == tag_maintain:
			tag_delete = merge_object.tag_id_2
		else:
			tag_delete = merge_object.tag_id_1

		tag_delete_datasets = p.toolkit.get_action('tag_show')({},{'id' : tag_delete, 'include_datasets': True})

		for ds in tag_delete_datasets['packages']:
			dataset = p.toolkit.get_action('package_show')({},{'id': ds['id'] })
			dataset['tags'].append(p.toolkit.get_action('tag_show')({},{'id':tag_maintain}))
			p.toolkit.get_action('package_update')({},dataset)
		   
		merge_objects = db.TagMergeSuggestion.by_tag_id(tag_delete)
		for m in merge_objects:
			m.delete()
			m.commit()

		p.toolkit.get_action('tag_delete')({},{'id': tag_delete})
		return 

	def save_merge_suggestions(self,suggestion_type='all'):

		suggestion_type = request.POST['method']

		if suggestion_type == 'all':
			suggestions = self.get_merge_suggestions(0);
			for s in suggestions:
				session = db.TagMergeSuggestion(s[0],s[1],0) 
				session.save()
			suggestions = self.get_merge_suggestions(1);
			for s in suggestions:
				session = db.TagMergeSuggestion(s[0],s[1],1) 
				session.save()
			suggestions = self.get_merge_suggestions(2);
			for s in suggestions:
				session.db.TagMergeSuggestion(s[0],s[1],2)
				session.save()

		else:
			suggestions = self.get_merge_suggestions(suggestion_type);
			for s in suggestions:
				session = db.TagMergeSuggestion(s[0],s[1],suggestion_type)
				session.save()
	
		return render('tagmanager/index.html')


	def get_merge_suggestions(self,suggestion_type=0, limit=None):
		from nltk.corpus import wordnet as wn

		print suggestion_type
		tags = p.toolkit.get_action('tag_list')({},{'all_fields' : True})
		T = len(tags)
		merge_list = []
		for t in range(0,T-1):
			if ((suggestion_type == '0') or (suggestion_type == 'all')):
				for s in range(t+1, T-1):
					if unidecode(tags[s]['name'].lower()) == unidecode(tags[t]['name'].lower()):
						merge_list.append([tags[s]['id'],tags[t]['id']])

			if (suggestion_type == '1') or (suggestion_type == 'all'):
				stri = tags[t]['name'] 
				if ([int(stri[i]) for i in range(0,len(stri)) if stri[i].isdigit()] == []) and (len(stri) > 3):
					for s in range(t,T-1):
						strj = tags[s]['name']
						if (len(strj) > 3) and ([int(strj[i]) for i in range(0,len(strj)) if strj[i].isdigit()] == []):
							if unidecode(tags[s]['name'].lower()) != unidecode(tags[t]['name'].lower()):
								d = Levenshtein.distance(tags[t]['name'],tags[s]['name'])
								if d < 3:
									merge_list.append([tags[s]['id'],tags[t]['id']])

			if (suggestion_type == '2') or (suggestion_type == 'all'):
				stri = tags[t]['name'] 
				if ([int(stri[i]) for i in range(0,len(stri)) if stri[i].isdigit()] == []) and (len(stri) > 3):
					syn1=wn.synsets(stri)
					if syn1 != []:
						for s in range(t,T-1):
							strj = tags[s]['name']
							if (len(strj) > 3) and ([int(strj[i]) for i in range(0,len(strj)) if strj[i].isdigit()] == []):
								syn2=wn.synsets(strj)
								if syn2 != []:
									if unidecode(tags[s]['name'].lower()) != unidecode(tags[t]['name'].lower()):
										if Levenshtein.distance(tags[t]['name'],tags[s]['name']) > 3:
											b = max(syn2[i].wup_similarity(syn1[0]) for i in range(len(syn2)))
											if b >= 1:
												merge_list.append([tags[s]['id'],tags[t]['id']])
		
		return merge_list

